# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""HTML processor."""

import json
import os
import queue
import typing
from html.parser import HTMLParser

from .utils import SEP

HTMLAttr = typing.List[typing.Tuple[str, typing.Union[str, None]]]
PARENT_CSS_STYLE = 'word-break: keep-all; overflow-wrap: anywhere;'
with open(
    os.path.join(os.path.dirname(__file__), 'skip_nodes.json'),
    encoding='utf-8') as f:
  SKIP_NODES: typing.Set[str] = set(json.load(f))


class ElementState(object):
  """Represents the state for an element.

  Attributes:
    tag (str): The tag name.
    to_skip (bool): Whether the content should be skipped or not.
  """

  def __init__(self, tag: str, to_skip: bool) -> None:
    self.tag = tag
    self.to_skip = to_skip


class TextContentExtractor(HTMLParser):
  """An HTML parser to extract text content.

  Attributes:
    output (str): Accumulated text content.
  """
  output = ''

  def handle_data(self, data: str) -> None:
    self.output += data


class HTMLChunkResolver(HTMLParser):
  """An HTML parser to resolve the given HTML string and semantic chunks.

  Attributes:
    output (str): The HTML string to output.
  """
  output = ''

  def __init__(self, chunks: typing.List[str], separator: str):
    """Initializes the parser.

    Args:
      chunks (List[str]): The chunks to resolve.
      separator (str): The separator string.
    """
    HTMLParser.__init__(self)
    self.chunks_joined = SEP.join(chunks)
    self.separator = separator
    self.to_skip = False
    self.scan_index = 0
    self.element_stack: queue.LifoQueue[ElementState] = queue.LifoQueue()

  def handle_starttag(self, tag: str, attrs: HTMLAttr) -> None:
    attr_pairs = []
    for attr in attrs:
      if attr[1] is None:
        attr_pairs.append(' ' + attr[0])
      else:
        attr_pairs.append(' %s="%s"' % (attr[0], attr[1]))
    encoded_attrs = ''.join(attr_pairs)
    self.element_stack.put(ElementState(tag, self.to_skip))
    if tag.upper() in SKIP_NODES:
      if not self.to_skip and self.chunks_joined[self.scan_index] == SEP:
        self.scan_index += 1
        self.output += self.separator
      self.to_skip = True
    self.output += '<%s%s>' % (tag, encoded_attrs)

  def handle_endtag(self, tag: str) -> None:
    self.output += '</%s>' % (tag)
    while not self.element_stack.empty():
      state = self.element_stack.get_nowait()
      if state.tag == tag:
        self.to_skip = state.to_skip
        break
      # If the close tag doesn't match the open tag, remove it and keep looking.
      # This means that close tags close their corresponding open tags.
      # e.g., `<span>abc<img>def</span>` or `<p>abc<span>def</p>` are both valid
      # HTML as per the HTML spec.
      # Note the HTML "adoption agency algorithm" isn't fully supported.
      # See https://html.spec.whatwg.org/multipage/parsing.html#an-introduction-to-error-handling-and-strange-cases-in-the-parser

  def handle_data(self, data: str) -> None:
    for char in data:
      if not char == self.chunks_joined[self.scan_index]:
        if not self.to_skip:
          self.output += self.separator
        self.scan_index += 1
      self.output += char
      self.scan_index += 1


def get_text(html: str) -> str:
  """Gets the text content from the input HTML string.

  Args:
    html (str): Input HTML string.

  Returns:
    The text content.
  """
  text_content_extractor = TextContentExtractor()
  text_content_extractor.feed(html)
  return text_content_extractor.output


def resolve(phrases: typing.List[str],
            html: str,
            separator: str = '\u200b') -> str:
  """Wraps phrases in the HTML string with non-breaking markup.

  Args:
    phrases (List[str]): The phrases included in the HTML string.
    html (str): The HTML string to resolve.
    separator (str, optional): The separator string.

  Returns:
    The HTML string with phrases wrapped in non-breaking markup.
  """
  resolver = HTMLChunkResolver(phrases, separator)
  resolver.feed(html)
  result = '<span style="%s">%s</span>' % (PARENT_CSS_STYLE, resolver.output)
  return result



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quality regression test."""

import os
import sys
import unittest

# module hack
LIB_PATH = os.path.join(os.path.dirname(__file__), '..')
sys.path.insert(0, os.path.abspath(LIB_PATH))

from budoux import load_default_japanese_parser, utils  # noqa (module hack)


class TestQuality(unittest.TestCase):

  def test_ja(self) -> None:
    errors = []
    parser = load_default_japanese_parser()
    fp = os.path.join(os.path.dirname(__file__), 'quality', 'ja.tsv')
    with open(fp, 'r', encoding='utf-8') as f:
      data = [line.split('\t') for line in f.readlines() if line[0] != '#']
    expected_sentences = [line[1].strip() for line in data if len(line) > 1]
    for expected in expected_sentences:
      result = utils.SEP.join(parser.parse(expected.replace(utils.SEP, '')))
      if result != expected:
        errors.append((expected, result))
    self.assertEqual(
        len(errors), 0, 'Failing sentences:\n{}'.format('\n'.join(
            [f'expected:{err[0]}\tactual:{err[1]}' for err in errors])))



#!/usr/bin/env python
# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BudouX Script to provide CLI for user."""
import argparse
import json
import os
import shutil
import sys
import textwrap
import typing
from importlib import resources

import budoux

ArgList = typing.Optional[typing.List[str]]
# Using typing.Any for resource related objects to avoid complex conditional
# imports for Traversable which varies across Python versions.
models: typing.Any = resources.files('budoux') / "models"
langs: typing.Dict[str, typing.Any] = {
    model.name[:-5]: model
    for model in models.iterdir()
    if model.name.endswith(".json")
}


class BudouxHelpFormatter(argparse.ArgumentDefaultsHelpFormatter,
                          argparse.RawDescriptionHelpFormatter):
  pass


def check_file(path: str) -> str:
  """Check if a given filepath exists or not.

  Args:
      path (str): Model path

  Raises:
      FileNotFoundError: Raise if given path does not exist.

  Returns:
      str: A model path.
  """
  if os.path.isfile(path):
    return path
  else:
    raise argparse.ArgumentTypeError(f"'{path}' is not found.")


def check_lang(lang: str) -> typing.Any:
  """Check if given language exists or not.

  Args:
      lang (str): language code (e.g.: 'ja')

  Raises:
      argparse.ArgumentTypeError: Raise if no model for given language exists.

  Returns:
      The model path.
  """
  if lang in langs:
    return langs[lang]
  else:
    raise argparse.ArgumentTypeError(
        f"'{lang}' does not exist in builtin models. (supported languages: {list(langs.keys())})"
    )


def parse_args(test: ArgList = None) -> argparse.Namespace:
  """Parse commandline arguments.

  Args:
      test (typing.Optional[typing.List[str]], optional): Commandline args for testing. Defaults to None.

  Returns:
      argparse.Namespace: Parsed data of args.
  """
  parser = argparse.ArgumentParser(
      prog="budoux",
      formatter_class=(lambda prog: BudouxHelpFormatter(
          prog,
          **{
              "width": shutil.get_terminal_size(fallback=(120, 50)).columns,
              "max_help_position": 30,
          },
      )),
      description=textwrap.dedent("""\
        BudouX is the successor to Budou,
        the machine learning powered line break organizer tool."""),
      epilog="\n- ".join(
          ["supported languages of `-l`, `--lang`:", *langs.keys()]))

  parser.add_argument("text", metavar="TXT", nargs="?", type=str, help="text")
  parser.add_argument(
      "-H",
      "--html",
      action="store_true",
      help="HTML mode",
  )
  model_select_group = parser.add_mutually_exclusive_group()
  model_select_group.add_argument(
      "-m",
      "--model",
      metavar="JSON",
      type=check_file,
      default=check_lang('ja'),
      help="custom model file path",
  )
  model_select_group.add_argument(
      "-l",
      "--lang",
      metavar="LANG",
      type=check_lang,
      help="language of custom model",
  )
  parser.add_argument(
      "-s",
      "--sep",
      metavar="STR",
      type=str,
      default="\n",
      help="output phrase separator in TEXT mode",
  )
  parser.add_argument(
      "-d",
      "--delim",
      metavar="STR",
      type=str,
      default="---",
      help="output sentence delimiter in TEXT mode",
  )
  parser.add_argument(
      "-V",
      "--version",
      action="version",
      version="%(prog)s {}".format(budoux.__version__),
  )
  if test is not None:
    return parser.parse_args(test)
  else:
    return parser.parse_args()


def _main(test: ArgList = None) -> str:
  args = parse_args(test=test)
  model_path = args.lang or args.model
  # Using open() directly assuming model_path is a path-like object.
  with open(model_path, 'r', encoding='utf-8') as f:
    model = json.load(f)

  parser = budoux.Parser(model)
  if args.html:
    if args.text is None:
      inputs_html = sys.stdin.read()
    else:
      inputs_html = args.text
    res = parser.translate_html_string(inputs_html)
  else:
    if args.text is None:
      inputs = [v.rstrip() for v in sys.stdin.readlines()]
    else:
      inputs = [v.rstrip() for v in args.text.splitlines()]
    outputs = [parser.parse(sentence) for sentence in inputs]
    combined_output = [args.sep.join(output) for output in outputs]
    ors = "\n" + args.delim + "\n"
    res = ors.join(combined_output)

  return res


def main(test: ArgList = None) -> None:
  try:
    print(_main(test))
  except KeyboardInterrupt:
    exit(0)


if __name__ == "__main__":
  main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Tests the HTML Processor."

import os
import sys
import unittest

# module hack
LIB_PATH = os.path.join(os.path.dirname(__file__), '..')
sys.path.insert(0, os.path.abspath(LIB_PATH))

from budoux import html_processor  # noqa (module hack)


class TestTextContentExtractor(unittest.TestCase):

  def test_output(self) -> None:
    input = '<p><a href="#">Hello</a>, <b>World</b></p>'
    expected = 'Hello, World'
    extractor = html_processor.TextContentExtractor()
    extractor.feed(input)
    self.assertEqual(
        extractor.output, expected,
        'Text content should be extacted from the given HTML string.')


class TestHTMLChunkResolver(unittest.TestCase):

  def test_output(self) -> None:
    input = '<p>ab<b>cde</b>f</p>'
    expected = '<p>ab<b>c<wbr>de</b>f</p>'
    resolver = html_processor.HTMLChunkResolver(['abc', 'def'], '<wbr>')
    resolver.feed(input)
    self.assertEqual(resolver.output, expected,
                     'WBR tags should be inserted as specified by chunks.')

  def test_unpaired(self) -> None:
    input = '<p>abcdef</p></p>'
    expected = '<p>abc<wbr>def</p></p>'
    resolver = html_processor.HTMLChunkResolver(['abc', 'def'], '<wbr>')
    resolver.feed(input)
    self.assertEqual(resolver.output, expected,
                     'Unpaired close tag should not cause errors.')

  def test_nobr(self) -> None:
    input = '<p>ab<nobr>cde</nobr>f</p>'
    expected = '<p>ab<nobr>cde</nobr>f</p>'
    resolver = html_processor.HTMLChunkResolver(['abc', 'def'], '<wbr>')
    resolver.feed(input)
    self.assertEqual(resolver.output, expected,
                     'WBR tags should not be inserted if in NOBR.')

  def test_after_nobr(self) -> None:
    input = '<p>ab<nobr>xy</nobr>abcdef</p>'
    expected = '<p>ab<nobr>xy</nobr>abc<wbr>def</p>'
    resolver = html_processor.HTMLChunkResolver(['abxyabc', 'def'], '<wbr>')
    resolver.feed(input)
    self.assertEqual(resolver.output, expected,
                     'WBR tags should be inserted if after NOBR.')

  def test_img_in_nobr(self) -> None:
    input = '<p>ab<nobr>x<img>y</nobr>abcdef</p>'
    expected = '<p>ab<nobr>x<img>y</nobr>abc<wbr>def</p>'
    resolver = html_processor.HTMLChunkResolver(['abxyabc', 'def'], '<wbr>')
    resolver.feed(input)
    self.assertEqual(resolver.output, expected,
                     'IMG should not affect surrounding NOBR.')


class TestResolve(unittest.TestCase):

  def test_with_simple_text_input(self) -> None:
    chunks = ['abc', 'def']
    html = 'abcdef'
    result = html_processor.resolve(chunks, html)
    expected = '<span style="word-break: keep-all; overflow-wrap: anywhere;">abc\u200bdef</span>'
    self.assertEqual(result, expected)

  def test_with_standard_html_input(self) -> None:
    chunks = ['abc', 'def']
    html = 'ab<a href="http://example.com">cd</a>ef'
    result = html_processor.resolve(chunks, html)
    expected = '<span style="word-break: keep-all; overflow-wrap: anywhere;">ab<a href="http://example.com">c\u200bd</a>ef</span>'
    self.assertEqual(result, expected)

  def test_with_nodes_to_skip(self) -> None:
    chunks = ['abc', 'def', 'ghi']
    html = "a<button>bcde</button>fghi"
    result = html_processor.resolve(chunks, html)
    expected = '<span style="word-break: keep-all; overflow-wrap: anywhere;">a<button>bcde</button>f\u200bghi</span>'
    self.assertEqual(result, expected)

  def test_with_break_before_skip(self) -> None:
    chunks = ['abc', 'def', 'ghi', 'jkl']
    html = "abc<button>defghi</button>jkl"
    result = html_processor.resolve(chunks, html)
    expected = '<span style="word-break: keep-all; overflow-wrap: anywhere;">abc\u200b<button>defghi</button>\u200bjkl</span>'
    self.assertEqual(result, expected)

  def test_with_nothing_to_split(self) -> None:
    chunks = ['abcdef']
    html = 'abcdef'
    result = html_processor.resolve(chunks, html)
    expected = '<span style="word-break: keep-all; overflow-wrap: anywhere;">abcdef</span>'
    self.assertEqual(result, expected)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests the BudouX CLI."""

import io
import sys
import unittest
from os.path import abspath, dirname, join

# module hack
LIB_PATH = join(dirname(__file__), '..')
sys.path.insert(0, abspath(LIB_PATH))

from budoux import main  # noqa (module hack)

if isinstance(sys.stdin, io.TextIOWrapper):
  sys.stdin.reconfigure(encoding='utf-8')

if isinstance(sys.stdout, io.TextIOWrapper):
  sys.stdout.reconfigure(encoding='utf-8')


class TestCommonOption(unittest.TestCase):

  def test_cmdargs_invalid_option(self) -> None:
    cmdargs = ['-v']
    with self.assertRaises(SystemExit) as cm:
      main.parse_args(cmdargs)

    self.assertEqual(cm.exception.code, 2)

  def test_cmdargs_help(self) -> None:
    cmdargs = ['-h']
    with self.assertRaises(SystemExit) as cm:
      main.parse_args(cmdargs)

    self.assertEqual(cm.exception.code, 0)

  def test_cmdargs_version(self) -> None:
    cmdargs = ['-V']
    with self.assertRaises(SystemExit) as cm:
      main.parse_args(cmdargs)

    self.assertEqual(cm.exception.code, 0)


class TestModelOption(unittest.TestCase):

  def test_cmdargs_invalid_json(self) -> None:
    cmdargs = ['-m', '404.json']
    with self.assertRaises(SystemExit) as cm:
      main.parse_args(cmdargs)

    self.assertEqual(cm.exception.code, 2)

  def test_cmdargs_invalid_lang_1(self) -> None:
    cmdargs = ['-l', 'aa']
    with self.assertRaises(SystemExit) as cm:
      main.parse_args(cmdargs)

    self.assertEqual(cm.exception.code, 2)

  def test_cmdargs_invalid_lang_2(self) -> None:
    cmdargs = ['-l', 'ja-abc']
    with self.assertRaises(SystemExit) as cm:
      main.parse_args(cmdargs)

    self.assertEqual(cm.exception.code, 2)

  def test_cmdargs_lang_ja(self) -> None:
    cmdargs = ['-l', 'ja', '今日は良い天気ですね。']
    output = main._main(cmdargs)

    self.assertEqual(output, '今日は\n良い\n天気ですね。')

  def test_cmdargs_lang_zh_hans(self) -> None:
    cmdargs = ['-l', 'zh-hans', '今天天气晴朗。']
    output = main._main(cmdargs)

    self.assertEqual(output, '今天\n天气\n晴朗。')


class TestTextArguments(unittest.TestCase):

  def test_cmdargs_single_text(self) -> None:
    cmdargs = ['これはテストです。']
    output = main._main(cmdargs)

    self.assertEqual(output, "これは\nテストです。")

  def test_cmdargs_single_multiline_text(self) -> None:
    cmdargs = ["これはテストです。\n今日は晴天です。"]
    output = main._main(cmdargs)

    self.assertEqual(output, "これは\nテストです。\n---\n今日は\n晴天です。")

  def test_cmdargs_single_multiline_text_with_delimiter(self) -> None:
    cmdargs = ["これはテストです。\n今日は晴天です。", "-d", "@"]
    output = main._main(cmdargs)

    self.assertEqual(output, "これは\nテストです。\n@\n今日は\n晴天です。")

  def test_cmdargs_single_multiline_text_with_empty_delimiter(self) -> None:
    cmdargs = ["これはテストです。\n今日は晴天です。", "-d", ""]
    output = main._main(cmdargs)

    self.assertEqual(output, "これは\nテストです。\n\n今日は\n晴天です。")

  def test_cmdargs_multi_text(self) -> None:
    cmdargs = ['これはテストです。', '今日は晴天です。']
    with self.assertRaises(SystemExit) as cm:
      main.main(cmdargs)

    self.assertEqual(cm.exception.code, 2)

  def test_cmdargs_single_html(self) -> None:
    cmdargs = ['-H', '今日は<b>とても天気</b>です。']
    output = main._main(cmdargs)

    self.assertEqual(
        output, '<span style="word-break: keep-all; overflow-wrap: anywhere;">'
        '今日は<b>\u200bとても\u200b天気</b>です。</span>')

  def test_cmdargs_multi_html(self) -> None:
    cmdargs = ['-H', '今日は<b>とても天気</b>です。', 'これは<b>テスト</b>です。']
    with self.assertRaises(SystemExit) as cm:
      main._main(cmdargs)

    self.assertEqual(cm.exception.code, 2)


class TestStdin(unittest.TestCase):

  def test_cmdargs_blank_stdin(self) -> None:
    with open(
        join(abspath(dirname(__file__)), "in/1.in"),
        "r",
        encoding=sys.getdefaultencoding()) as f:
      sys.stdin = f
      output = main._main([])

    self.assertEqual(output, "")

  def test_cmdargs_text_stdin(self) -> None:
    with open(
        join(abspath(dirname(__file__)), "in/2.in"),
        "r",
        encoding=sys.getdefaultencoding()) as f:
      sys.stdin = f
      output = main._main([])

    self.assertEqual(output, "これは\nテストです。")

  def test_cmdargs_html_stdin(self) -> None:
    with open(
        join(abspath(dirname(__file__)), "in/3.in"),
        "r",
        encoding=sys.getdefaultencoding()) as f:
      sys.stdin = f
      output = main._main(["-H"])

    self.assertEqual(
        output, '<span style="word-break: keep-all; overflow-wrap: anywhere;">'
        'これは<b>\u200bテスト</b>です。\u200b\n'
        '</span>')


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests the BudouX parser."""

import os
import sys
import unittest

# module hack
LIB_PATH = os.path.join(os.path.dirname(__file__), '..')
sys.path.insert(0, os.path.abspath(LIB_PATH))

from budoux import parser  # noqa (module hack)


class TestParser(unittest.TestCase):
  TEST_SENTENCE = 'abcdeabcd'

  def test_parse(self) -> None:
    p = parser.Parser({
        'UW4': {
            'a': 10000
        },  # means "should separate right before 'a'".
    })
    chunks = p.parse(TestParser.TEST_SENTENCE)
    self.assertListEqual(chunks, ['abcde', 'abcd'],
                         'Should separate if a strong feature item supports.')

    p = parser.Parser({
        'UW4': {
            'b': 10000
        },  # means "should separate right before 'b'".
    })
    chunks = p.parse(TestParser.TEST_SENTENCE)
    self.assertListEqual(
        chunks, ['a', 'bcdea', 'bcd'],
        'Should separate even if it makes the first character a sole phrase.')

    p = parser.Parser({})
    chunks = p.parse('')
    self.assertListEqual(chunks, [],
                         'Should return a blank list when the input is blank.')

  def test_translate_html_string(self) -> None:
    p = parser.Parser({
        'UW4': {
            'a': 10000
        },  # means "should separate right before 'a'".
    })

    input_html = 'xyzabcd'
    expected_html = (
        '<span style="word-break: keep-all; overflow-wrap: anywhere;">'
        'xyz\u200babcd</span>')
    output_html = p.translate_html_string(input_html)
    self.assertEqual(
        output_html, expected_html,
        'Should output a html string with a SPAN parent with proper style attributes.'
    )

    input_html = 'xyz<script>alert(1);</script>xyzabc'
    # TODO: Because the content for skip elements are included, this test tries
    # to break before "alert". We may want to distinguish "skip from the
    # content" and "skip breaking" in future.
    expected_html = (
        '<span style="word-break: keep-all; overflow-wrap: anywhere;">'
        'xyz\u200b<script>alert(1);</script>xyz\u200babc</span>')
    output_html = p.translate_html_string(input_html)
    self.assertEqual(output_html, expected_html,
                     'Should pass script tags as is.')

    input_html = 'xyz<code>abc</code>abc'
    expected_html = (
        '<span style="word-break: keep-all; overflow-wrap: anywhere;">'
        'xyz\u200b<code>abc</code>\u200babc</span>')
    output_html = p.translate_html_string(input_html)
    self.assertEqual(output_html, expected_html,
                     'Should skip some specific tags.')

    input_html = 'xyza<a href="#" hidden>bc</a>abc'
    expected_html = (
        '<span style="word-break: keep-all; overflow-wrap: anywhere;">'
        'xyz\u200ba<a href="#" hidden>bc</a>\u200babc</span>')
    output_html = p.translate_html_string(input_html)
    self.assertEqual(output_html, expected_html,
                     'Should not ruin attributes of child elements.')

    input_html = 'xyza🇯🇵🇵🇹abc'
    expected_html = (
        '<span style="word-break: keep-all; overflow-wrap: anywhere;">'
        'xyz\u200ba🇯🇵🇵🇹\u200babc</span>')
    output_html = p.translate_html_string(input_html)
    self.assertEqual(output_html, expected_html, 'Should work with emojis.')


class TestDefaultParser(unittest.TestCase):

  def test_load_default_japanese_parser(self) -> None:
    p_ja = parser.load_default_japanese_parser()
    phrases = p_ja.parse('Google の使命は、世界中の情報を整理し、世界中の人がアクセスできて使えるようにすることです。')
    self.assertListEqual(phrases, [
        'Google の',
        '使命は、',
        '世界中の',
        '情報を',
        '整理し、',
        '世界中の',
        '人が',
        'アクセスできて',
        '使えるように',
        'する',
        'ことです。',
    ])

  def test_load_default_simplified_chinese_parser(self) -> None:
    p_hans = parser.load_default_simplified_chinese_parser()
    phrases = p_hans.parse('我们的使命是整合全球信息，供大众使用，让人人受益。')
    self.assertListEqual(phrases, [
        '我们',
        '的',
        '使命',
        '是',
        '整合',
        '全球',
        '信息，',
        '供',
        '大众',
        '使用，',
        '让',
        '人',
        '人',
        '受益。',
    ])

  def test_load_default_traditional_chinese_parser(self) -> None:
    p_hant = parser.load_default_traditional_chinese_parser()
    phrases = p_hant.parse('我們的使命是匯整全球資訊，供大眾使用，使人人受惠。')
    self.assertListEqual(phrases, [
        '我們',
        '的',
        '使命',
        '是',
        '匯整',
        '全球',
        '資訊，',
        '供',
        '大眾',
        '使用，',
        '使',
        '人',
        '人',
        '受惠。',
    ])


if __name__ == '__main__':
  unittest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BudouX parser to provide semantic chunks."""

import json
import os
import typing

from .html_processor import get_text, resolve

MODEL_DIR = os.path.join(os.path.dirname(__file__), 'models')


class Parser:
  """BudouX's Parser.

  The main parser object with a variety of class methods to provide semantic
  chunks and markups from the given input string.

  Attributes:
    _model: A dict mapping a feature (str) and its score (int).
  """

  def __init__(self, model: typing.Dict[str, typing.Dict[str, int]]):
    """Initializes the parser.

    Args:
      model (Dict[str, Dict[str, int]]): A dict mapping a feature and its score.
    """
    self._model = model
    self._base_score = -sum(sum(g.values()) for g in self._model.values()) * 0.5

  @property
  def model(self) -> typing.Dict[str, typing.Dict[str, int]]:
    return self._model

  def parse(self, sentence: str) -> typing.List[str]:
    """Parses the input sentence and returns a list of semantic chunks.

    Args:
      sentence (str): An input sentence.

    Returns:
      A list of semantic chunks (List[str]).
    """
    if sentence == '':
      return []
    chunks = [sentence[0]]
    for i in range(1, len(sentence)):
      score = self._base_score
      if i > 2:
        score += self._model.get('UW1', {}).get(sentence[i - 3], 0)
      if i > 1:
        score += self._model.get('UW2', {}).get(sentence[i - 2], 0)
      score += self._model.get('UW3', {}).get(sentence[i - 1], 0)
      score += self._model.get('UW4', {}).get(sentence[i], 0)
      if i + 1 < len(sentence):
        score += self._model.get('UW5', {}).get(sentence[i + 1], 0)
      if i + 2 < len(sentence):
        score += self._model.get('UW6', {}).get(sentence[i + 2], 0)

      if i > 1:
        score += self._model.get('BW1', {}).get(sentence[i - 2:i], 0)
      score += self._model.get('BW2', {}).get(sentence[i - 1:i + 1], 0)
      if i + 1 < len(sentence):
        score += self._model.get('BW3', {}).get(sentence[i:i + 2], 0)

      if i > 2:
        score += self._model.get('TW1', {}).get(sentence[i - 3:i], 0)
      if i > 1:
        score += self._model.get('TW2', {}).get(sentence[i - 2:i + 1], 0)
      if i + 1 < len(sentence):
        score += self._model.get('TW3', {}).get(sentence[i - 1:i + 2], 0)
      if i + 2 < len(sentence):
        score += self._model.get('TW4', {}).get(sentence[i:i + 3], 0)

      if score > 0:
        chunks.append(sentence[i])
      else:
        chunks[-1] += sentence[i]
    return chunks

  def translate_html_string(self, html: str) -> str:
    """Translates the given HTML string with markups for semantic line breaks.

    Args:
      html (str): An input html string.

    Returns:
      The translated HTML string (str).
    """
    # TODO: Align with the JavaScript API regarding the parent element addition.
    text_content = get_text(html)
    chunks = self.parse(text_content)
    return resolve(chunks, html)


def load_default_japanese_parser() -> Parser:
  """Loads a parser equipped with the default Japanese model.

  Returns:
    A parser (:obj:`budoux.Parser`).
  """
  with open(os.path.join(MODEL_DIR, 'ja.json'), encoding='utf-8') as f:
    model = json.load(f)
  return Parser(model)


def load_default_simplified_chinese_parser() -> Parser:
  """Loads a parser equipped with the default Simplified Chinese model.

  Returns:
    A parser (:obj:`budoux.Parser`).
  """
  with open(os.path.join(MODEL_DIR, 'zh-hans.json'), encoding='utf-8') as f:
    model = json.load(f)
  return Parser(model)


def load_default_traditional_chinese_parser() -> Parser:
  """Loads a parser equipped with the default Traditional Chinese model.

  Returns:
    A parser (:obj:`budoux.Parser`).
  """
  with open(os.path.join(MODEL_DIR, 'zh-hant.json'), encoding='utf-8') as f:
    model = json.load(f)
  return Parser(model)


def load_default_thai_parser() -> Parser:
  """Loads a parser equipped with the default Thai model.

  Returns:
    A parser (:obj:`budoux.Parser`).
  """
  with open(os.path.join(MODEL_DIR, 'th.json'), encoding='utf-8') as f:
    model = json.load(f)
  return Parser(model)
